/* -*-c-*-
 *
 * $Id: f-binpoly.c,v 1.8 2004/04/02 01:03:49 mdw Exp $
 *
 * Binary fields with polynomial basis representation
 *
 * (c) 2004 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: f-binpoly.c,v $
 * Revision 1.8  2004/04/02 01:03:49  mdw
 * Miscellaneous constification.
 *
 * Revision 1.7  2004/04/01 21:28:41  mdw
 * Normal basis support (translates to poly basis internally).  Rewrite
 * EC and prime group table generators in awk, so that they can reuse data
 * for repeated constants.
 *
 * Revision 1.6  2004/04/01 12:50:09  mdw
 * Add cyclic group abstraction, with test code.  Separate off exponentation
 * functions for better static linking.  Fix a buttload of bugs on the way.
 * Generally ensure that negative exponents do inversion correctly.  Add
 * table of standard prime-field subgroups.  (Binary field subgroups are
 * currently unimplemented but easy to add if anyone ever finds a good one.)
 *
 * Revision 1.5  2004/03/27 17:54:11  mdw
 * Standard curves and curve checking.
 *
 * Revision 1.4  2004/03/23 15:19:32  mdw
 * Test elliptic curves more thoroughly.
 *
 * Revision 1.3  2004/03/23 12:08:26  mdw
 * Random field-element selection.
 *
 * Revision 1.2  2004/03/21 22:52:06  mdw
 * Merge and close elliptic curve branch.
 *
 * Revision 1.1.2.1  2004/03/21 22:39:46  mdw
 * Elliptic curves on binary fields work.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include <mLib/sub.h>

#include "field.h"
#include "gf.h"
#include "gfreduce.h"
#include "mprand.h"
#include "gfn.h"

/*----- Polynomial basis --------------------------------------------------*/

typedef struct fctx {
  field f;
  gfreduce r;
} fctx;

/* --- Field operations --- */

static void fdestroy(field *ff)
  { fctx *f = (fctx *)ff; gfreduce_destroy(&f->r); DESTROY(f); }

static mp *frand(field *f, mp *d, grand *r)
  { return (mprand(d, f->nbits, r, 0)); }

static int fzerop(field *ff, mp *x) { return (!MP_LEN(x)); }

static mp *fadd(field *ff, mp *d, mp *x, mp *y) { return (gf_add(d, x, y)); }

static mp *fmul(field *ff, mp *d, mp *x, mp *y) {
  fctx *f = (fctx *)ff; d = gf_mul(d, x, y);
  return (gfreduce_do(&f->r, d, d));
}

static mp *fsqr(field *ff, mp *d, mp *x) {
  fctx *f = (fctx *)ff; d = gf_sqr(d, x);
  return (gfreduce_do(&f->r, d, d));
}

static mp *finv(field *ff, mp *d, mp *x)
  { fctx *f = (fctx *)ff; gf_gcd(0, 0, &d, f->r.p, x); return (d); }

static mp *freduce(field *ff, mp *d, mp *x)
  { fctx *f = (fctx *)ff; return (gfreduce_do(&f->r, d, x)); }

static mp *fsqrt(field *ff, mp *d, mp *x)
  { fctx *f = (fctx *)ff; return (gfreduce_sqrt(&f->r, d, x)); }

static mp *fquadsolve(field *ff, mp *d, mp *x)
  { fctx *f = (fctx *)ff; return (gfreduce_quadsolve(&f->r, d, x)); }

/* --- Field operations table --- */

static const field_ops fops = {
  FTY_BINARY, "binpoly",
  fdestroy, frand, field_stdsamep,
  freduce, field_id,
  fzerop, field_id, fadd, fadd, fmul, fsqr, finv, freduce, fsqrt,
  fquadsolve,
  0, 0, 0, 0
};

/* --- @field_binpoly@ --- *
 *
 * Arguments:	@mp *p@ = the reduction polynomial
 *
 * Returns:	A pointer to the field.
 *
 * Use:		Creates a field structure for a binary field mod @p@.
 */

field *field_binpoly(mp *p)
{
  fctx *f = CREATE(fctx);
  f->f.ops = &fops;
  f->f.zero = MP_ZERO;
  f->f.one = MP_ONE;
  f->f.nbits = mp_bits(p) - 1;
  f->f.noctets = (f->f.nbits + 7) >> 3;
  gfreduce_create(&f->r, p);
  f->f.m = f->r.p;
  return (&f->f);
}

/*----- Normal basis ------------------------------------------------------*/

typedef struct fnctx {
  fctx f;
  gfn ntop, pton;
} fnctx;

/* --- Field operations --- */

static void fndestroy(field *ff) {
  fnctx *f = (fnctx *)ff; gfreduce_destroy(&f->f.r);
  gfn_destroy(&f->ntop); gfn_destroy(&f->pton);
  DESTROY(f);
}

static int fnsamep(field *ff, field *gg) {
  fnctx *f = (fnctx *)ff, *g = (fnctx *)gg;
  return (MP_EQ(f->ntop.r[0], g->ntop.r[0]) && field_stdsamep(ff, gg));
}

static mp *fnin(field *ff, mp *d, mp *x)
  { fnctx *f = (fnctx *)ff; return (gfn_transform(&f->ntop, d, x)); }

static mp *fnout(field *ff, mp *d, mp *x)
  { fnctx *f = (fnctx *)ff; return (gfn_transform(&f->pton, d, x)); }

/* --- Field operations table --- */

static const field_ops fnops = {
  FTY_BINARY, "binnorm",
  fndestroy, frand, fnsamep,
  fnin, fnout,
  fzerop, field_id, fadd, fadd, fmul, fsqr, finv, freduce, fsqrt,
  fquadsolve,
  0, 0, 0, 0
};

/* --- @field_binnorm@ --- *
 *
 * Arguments:	@mp *p@ = the reduction polynomial
 *		@mp *beta@ = representation of normal point
 *
 * Returns:	A pointer to the field.
 *
 * Use:		Creates a field structure for a binary field mod @p@ which
 *		uses a normal basis representation externally.  Computations
 *		are still done on a polynomial-basis representation.
 */

field *field_binnorm(mp *p, mp *beta)
{
  fnctx *f = CREATE(fnctx);
  f->f.f.ops = &fnops;
  f->f.f.zero = MP_ZERO;
  f->f.f.one = MP_ONE;
  f->f.f.nbits = mp_bits(p) - 1;
  f->f.f.noctets = (f->f.f.nbits + 7) >> 3;
  gfreduce_create(&f->f.r, p);
  f->f.f.m = f->f.r.p;
  gfn_create(p, beta, &f->ntop, &f->pton);
  return (&f->f.f);
}

/*----- That's all, folks -------------------------------------------------*/
